/***********************************************************************/
/*            2009 - Serkan Eryilmaz                                   */
/*            Code libre de droits                                     */
/***********************************************************************/

#ifndef KLOTSKI_H_INCLUDED
#define KLOTSKI_H_INCLUDED

#include <map>
#include <boost/array.hpp>
#include "Solver.h"

#define ROW 5
#define COL 4

namespace Klotski
{
    class KlotskiState : public Solver::State< KlotskiState >
    {
    private:

        /*
        Cette énumération nous définit les différentes directions vers lequels
        nous pouvons déplacer une pièce d'un état Klotski.
        */
        enum Directions
        {
            LEFT,
            RIGHT,
            UP,
            DOWN
        };

        /*
        Cette énumération définit les types de pièces dont peut disposer un
        plateau de jeu Klotski
        */
        enum Type
        {
            SQUARE,
            VOID,
            HORIZONTAL,
            VERTICAL,
            KLOTSKI
        };

        /*
        Cette structure définit une position. Nous devons pouvoir comparer différentes
        positions, donc nous surchargeons les opérateurs correspondants.
        */
        struct Position
        {
            unsigned short x;
            unsigned short y;

            Position( ) :
                    x( 0 ),
                    y( 0 )
            {
            }

            Position( const unsigned short& x_, const unsigned short& y_ ) :
                    x( x_ ),
                    y( y_ )
            {

            }

            Position( const Position &p ) :
                    x(p.x),
                    y(p.y)
            {
            }

            bool operator< ( const Position &p ) const
            {
                if ( x < p.x )
                {
                    return y <= p.y;
                }
                else
                {
                    return y < p.y;
                }

            }

            bool operator== ( const Position &p ) const
            {
                return x == p.x && y == p.y;
            }

            unsigned short Distance( const Position& p ) const
            {
                return std::max( x, p.x ) - std::min ( x, p.x ) + std::max( y, p.y ) - std::min( y, p.y );
            }
        };

        /*
        Cette structure permet de définir une pièce du plateau de jeu.
        Chaque pièce possède une position et un type
        */
        struct Piece
        {
            Position pos;   // Position de la pièce
            Type     type;  // Type de la pièce

            Piece( )
            {
            }

            Piece( const Piece& p ) :
                    pos( p.pos ),
                    type( p.type )
            {
            }

            Piece( const Position& p, const Type& t ) :
                    pos(p),
                    type(t)
            {
            }

            /*
            Une pièce est équivalente à une autre pièce si et seulement si sa position et son type
            sont équivalents.
            */
            bool operator==( const Piece& p ) const
            {
                return pos == p.pos && type == p.type;
            }
        };

        /*
        Notre conteneur de pièces
        La clé désigne l'identifiant (A,B,C,D,...) de la pièce
        */
        std::map<char, Piece> m_Board;

        /*
        Notre tableau de taille fixe qui facilitera l'affichage et la comparaison d'états de jeux.
        */
        boost::array< boost::array< char, COL >, ROW > m_cBoard;

        /*
        Un pointeur vers l'état mère, s'il y en a une.
        Ceci permettra également d'éviter de générer des états successeurs étants similaires
        à l'état mère.
        */
        KlotskiState *m_pParent;

        /*
        Cette variable contiendra le coût de l'état actuel, à partir de l'état initial.
        */
        int m_iG;

        /*
        Permettra de déterminer si 2 états de jeux sont exactement similaires.
        */
        bool isEqual( const KlotskiState& k ) const;

        /*
        Permettra de déterminer si 2 états de jeux sont des états "mirroirs".
        Il faut éliminer ces états étant donné que les états mirroirs sont des états dont le coût jusqu'à la
        solution finale sera exactement identique, donc la recherche exhaustive de deux états mirroirs est
        complètement inutile.
        */
        bool isEqualOpposite( const KlotskiState& k ) const;

        /*
        Permettra de peupler notre tableau m_cBoard
        */
        void MakeArray();

        /*
        Permet de déterminer si une position p est une case vide.
        */
        char isFree( const Position& p ) const;

        /*
        Permet de déplacer une pièce.
        piece    : permet de déterminer quelle pièce à bouger
        x,yoffset: le déplacement à effectuer
        piece_id : l'identifiant de la pièce à déplacer
        void_id  : la case vide qui est concernée par le déplacement -1 désigne le fait que les 2 cases sont concernées
        */
        void movePiece( const Piece& piece, const short& xoffset, const short& yoffset, const char& piece_id, const char& void_id );

        /*
        Détermine si une pièce piece peut être déplacer dans la direction indiquée.
        La valeur de retour est l'identifiant de la case vide concernée par le déplacement si le déplacement est possible et qu'il ne concerne qu'une seule case vide.

        Si 2 cases sont concernées, la valeur de retour sera de 1 ou de 0 selon la réponse.
        */
        char movePossible( const Piece& piece, unsigned short direction ) const;

    public:

        KlotskiState( );

        KlotskiState& operator=( const KlotskiState& k );

        KlotskiState *getParent();

        // Permet d'initialiser un état.
        // Cette méthode n'est à appeller que pour initialiser un état de départ.
        bool Init( const std::string& );

        // Retourne le coût pour atteindre l'état courant
        int getCost() const;

        // Retourne vrai si l'état courant est un état final
        bool isGoal() const;

        // Permet de comparer l'équivalence entre 2 états
        bool operator==( const KlotskiState& k ) const;

        // Permet d'afficher l'état courant à l'écran
        void Print( ) const;

        // Permet de retourner une liste d'états successeurs.
        const std::list< KlotskiState > getPossibleNextMoves( );
    };
}

#endif // KLOTSKI_H_INCLUDED
